/* 	Sql2ClassConsole.java
 *	Copyright (c) 	2001	   Dennis Groenendijk
 *					2002-2006, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: 21 mei 2001
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
* [When]			[Who]							[What]
* 06-06-2001   ratl 								Added catalog as option, user not longer required
* 28-08-2002   dennis@brains2b.nl		made initialization more public 
* 08-12-2002		"			"					closed connection when no longer needed.
* 08-03-2003 		"			" 				Improved displaying info an logging
* 04-11-2003		"			"					Implemented Main class and TypeWriter project
* 														Removed unnecessary accessors
* 22-03-2004	dennis@brains2b.nl		Added options from parameters
* 10-01-2005 	dennis@brains2b.nl		Added trim option
* 02-08-2006	dennis@brains2b.nl	    Option for setting the size of the Java types from Oracle types
* 08-09-2006	dennis@brains2b.nl	    Option for FIXED_CHAR on database type
* 28-11-2006    dennis@brains2b.nl  	#856 Use serialID instead of 0 for empty class
* 30-01-2007    dennis@brains2b.nl      #909 Implemented using default values
* 12-02-2007    dennis@brains2b.nl      serialID() should be used as function
* 17-01-2008	dennis@brains2b.nl      changes for new DataC version 
* TODO Javadoc for model option
* TODO support for classpath from Connector
*/
package org.brains2b.sql.toclass;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.brains2b.console.ArgumentException;
import org.brains2b.console.ArgumentMap;
import org.brains2b.console.Main;
import org.brains2b.data.DataException;
import org.brains2b.log.Logger;
import org.brains2b.sql.Connector;
import org.brains2b.sql.LocalConnector;
import org.brains2b.sql.meta.SQLMetaData;
import org.brains2b.sql.meta.TableModel;
import org.brains2b.sql.meta.TypesInfo;

/** Console Application for running Sql2Class
 *  <p>Sql2Class is a function which transforms a Database table into a java Class
 *  and writes a control class to interact with the database.
 *  <p>Sql2Class takes a number of parameters:
 *  <table><tr><th>parameter</th<th>short form</th><th>required</th><th>description</th></tr>
 *  <tr><td>driver</td><td>d</td><td>yes</td><td>qualified class name of the JDBC driver to use</td>
 *  <tr><td>url</td><td>r</td><td>yes</td><td>JDBC url to connect to the datasoource</td>
 *  <tr><td>user</td><td>u</td><td>no</td><td>username for the datasource, required only if JDBC connection requires it</td>
 *  <tr><td>password</td><td>p</td><td>no</td><td>password for the datasource, required only if JDBC connection requires it</td>
 *  <tr><td>table</td><td>t</td><td>yes</td><td>The table or view you want to generate</td>
 *  <tr><td>location</td><td>l</td><td>no</td><td>the path where to write the newly generated classes to. The package name
 * 	will be appended to the end of this location. <i>If no location is defined the directory where Sql2Class was started is used</td>
 *  <tr><td>package</td><td>g</td><td>no</td><td>The package for this class. This will be both set in de class files as used
 * to append to the path as directory to create the classes in.</td>
 *  <tr><td>readonly</td><td>w</td><td>no</td><td>Define that the control should be generated to as read-only</td></tr>
 *  <tr><td>optimistic</td><td>o</td><td>no</td><td>Use the OptimisticLocking and control code to identify the 
 *  data class as using optimistic locking. No locking mechanisme is providided if this is not set</td></tr>
 *   <tr><td>forcerowid</td><td>f</td><td>no</td><td>Use a provided unique row identification instead of primary keys,
 *  or unique indexes to generate the data control. If there are no primary keys or unique indexes this option will be 
 *  used by default.</td></tr>
 *  <p>Sql2Class can take the necessary parameters from a property file <code>sql2class.properties</code> or from the commandline.
 * The commandline overwrites parameters set in de sql2class.properties file. Use the parameter names to define the properties.
 *  <p>parameters passed to the commandline should be prefixed with <code>--</code> if you use the full name or <code>-</code>
 * if you use the short name.
 * <p>If the files you are generating  are the same as the files at the supplied location and package no new files
 * will be generated. Look at the <code>SqlClassWriter.isFileSame(File)</code> for the conditions.
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.42 [Jan 30, 2007]
 * @see SqlClassWriter#isFileSame(File)
 */
public class Sql2ClassConsole extends Main {

	public static final String VERSIONID="$1.01$";
	
	private final static String PROPERTIES = "sql2class.properties";

    /**
     * Constructor
     *
     */
    public Sql2ClassConsole() {
    	super();
    	init();
    }

	/**
	 * main
	 * @param args String[], the commandline arguments
	 */
    public static void main(String[] args) {
        try {
            Sql2ClassConsole sql = new Sql2ClassConsole();
            sql.initialize(args);
            sql.makeClass();
        } catch (Exception e) {
            e.printStackTrace();
            printHelp();
        }
    }

	/** initializes the commandline arguments for Sql2Class
	 * @see ArgumentMap
	 * @see Main
	 */ 
	private void init() {
		ArgumentMap map=new ArgumentMap();
		map.addOption("table",'t',true);
		map.addArgumentTo("table","tablename",true);
		map.addOption("driver",'d',true);
		map.addArgumentTo("driver","name",true);
		map.addOption("url",'r',true);
		map.addArgumentTo("url","name",true);
		map.addOption("user",'u',false);
		map.addArgumentTo("user","name",true);
		map.addOption("password",'p',false);
		map.addArgumentTo("password","word",true);
		map.addOption("catalog",'c',false);
		map.addArgumentTo("catalog","name",true);
		map.addOption("location",'l',false);
		map.addArgumentTo("location","dir",true);
		map.addOption("package",'g',false);
		map.addArgumentTo("package","name",true);
		map.addOption("readonly",'w',false);
		map.addOption("forcerowid",'f',false);
		map.addOption("optimistic",'o',false);
		map.addOption("trim",'o',false);
        map.addOption("default",'i',false);
		map.addOption("force",'0',false);
		map.addOption("model",'0',false);
		map.addArgumentTo("model","type",true);
		setArgumentMap(map);
	}

	/** initialize Sql2Class with the values from properties
	 * 
	 * @param prop Properties, the properties containing the arguments for Sql2Class
	 * @throws ArgumentException, thrown if a property from the file does not 
	 */
	protected void initialize(Properties prop) throws ArgumentException {
		String[] arguments = new String[prop.size()*2];
		int i=0;
		for(Iterator it=prop.keySet().iterator();it.hasNext();) {
			String key=(String) it.next();
			arguments[i++] = "--" + key;
			arguments[i++] = prop.getProperty(key);
		}
		makeArgumentList(arguments);
	}

    private void initialize(String[] args) throws Exception {
        Properties p=readProperties();
        String[] arguments = new String[args.length+p.size()*2];
        int i=0;
        for (Iterator it=p.keySet().iterator();it.hasNext();) {
        	String key=(String) it.next();
        	arguments[i++] = "--" + key;
        	arguments[i++] = p.getProperty(key);
        }
        System.arraycopy(args,0,arguments,p.size()*2,args.length);
        makeArgumentList(arguments);
    }

    protected void makeClass() throws SQLException,DataException {
    	
    	//make sure user name is in capitals
    	setArgument("user",getArgument("user").toUpperCase());
    	
        File fl_location = null;
        if (getArgument("location")!=null) {
            fl_location = new File(getArgument("location"));
        }
		Sql2File  s2f = new Sql2File(fl_location, getArgument("package"));
		Sql2DataC s2d = new Sql2DataC( fl_location, getArgument("package"));
		boolean s2fExists=false;
		boolean s2dExists=false;
        if (getArgument("force")==null) {
        	s2fExists = s2f.fileExists(getArgument("table"));
			s2dExists = s2d.fileExists(getArgument("table"));
			if (s2fExists && s2dExists) {
				return;
			}
        }
		
        Connector con = null;
        SQLMetaData meta = null;
        TableModel tm = null;
        
        try {
            // connect to the database
            con = new LocalConnector(
            	getArgument("driver"),
            	getArgument("url"),
            	getArgument("user"),
            	getArgument("password"));
            meta = new SQLMetaData(con.getConnection());
            meta.getTypeInfo().setDataModel(getModel());
            // get table information
            Logger.println("finding Table: " + getArgument("table"), Logger.DEBUG);
            List vt = meta.getTables(getArgument("catalog"), getArgument("user"), getArgument("table"));
            if (vt.size() != 1) {
                throw new DataException(
                    "No table found, or more then one with name " +getArgument("table"));
            } 
            tm = (TableModel) vt.get(0);
            
           
            Logger.println("finding Columns: " + getArgument("table"), Logger.DEBUG);
            
            meta.setColumns(tm);
           
        	Logger.println("finding PrimaryKeys: " + getArgument("table"), Logger.DEBUG);
        	meta.setPrimaryKey(tm, "true".equals(getArgument("forcerowid")));
      
        } finally {
        	con.close();
        }
		if (!s2fExists) {
			s2f.setTypeInfo(meta.getTypeInfo());
	        s2f.setTableModel(tm);
	        s2f.setTrimValue("true".equals(getArgument("trim")));
            s2f.setDefault("true".equals(getArgument("default")));
            
			s2f.setOptimisticLock("true".equals(getArgument("optimistic")));
    	    Thread t = new Thread(s2f);
        	t.run();
		}
		if (!s2dExists) {
			s2d.setTypeInfo(meta.getTypeInfo());
			s2d.setTableModel(tm);
			s2d.setReadOnly("true".equals(getArgument("readonly")));
			s2d.setDatabase(meta.getMetaData().getDatabaseProductName());
    	    Thread t = new Thread(s2d);
        	t.run();
		}

    }

    /**
     * read Properties
     * History
     * 21-05-2001 Ratl Reads properties direct and does not use ResourceBundle
     */
    private Properties readProperties() throws Exception {
		Properties pp = new Properties();
        try {
            File fl = new File(PROPERTIES);
            pp.load(new FileInputStream(fl));
        } catch (IOException ie) {
            //do nothing
        } catch (NullPointerException npex) {
            //do nothing
        } 
		return pp;
        
    }

    private int getModel() {
		if ("small".equals(getArgument("model"))) {
			return TypesInfo.SMALLEST;
		} else if ("large".equals(getArgument("model"))) {
			return TypesInfo.LARGE;
		} else {
			return TypesInfo.OPTIMIZE;
		}
    }

    private final static void printHelp() {
        //XXX On-screen help goes here
    }
  
}
